#pragma once

#include <tuple>
#include <type_traits>

#include <dune/common/hash.hh>

#include <amdis/common/IndexSeq.hpp>
#include <amdis/common/Mpl.hpp>
#include <amdis/common/Utility.hpp>

namespace AMDiS
{
  namespace Impl
  {
    // Recursive template code derived from Matthieu M.
    template <class Tuple, std::size_t I = std::tuple_size<Tuple>::value - 1>
    struct HashTupleImpl
    {
      static void apply(size_t& seed, Tuple const& tuple)
      {
        HashTupleImpl<Tuple, I-1>::apply(seed, tuple);
        Dune::hash_combine(seed, std::get<I>(tuple));
      }
    };

    template <class Tuple>
    struct HashTupleImpl<Tuple, 0>
    {
      static void apply(std::size_t& seed, Tuple const& tuple)
      {
        Dune::hash_combine(seed, std::get<0>(tuple));
      }
    };

  } // end namespace Impl

  namespace Impl
  {
    template <class Tuple, std::size_t N>
    struct ConstructTuple
    {
      // add arg to repeated constructor argument list
      template <class Arg, class... Args>
      static Tuple make(Arg&& arg, Args&&... args)
      {
        return ConstructTuple<Tuple, N-1>::make(
            std::forward<Arg>(arg), std::forward<Arg>(arg), std::forward<Args>(args)...);
      }
    };

    template <class Tuple>
    struct ConstructTuple<Tuple, 1>
    {
      template <class... Args>
      static Tuple make(Args&&... args)
      {
        static_assert(std::tuple_size<Tuple>::value == sizeof...(args),
            "Nr. of argument != tuple-size");
        return Tuple{std::forward<Args>(args)...};
      }
    };

    template <class Tuple>
    struct ConstructTuple<Tuple, 0>
    {
      template <class... Args>
      static Tuple make(Args&&... args)
      {
        static_assert(std::tuple_size<Tuple>::value == 0,
            "Construction of empty tuples with empty argument list only!");
        return Tuple{};
      }
    };


    template <class Tuple>
    struct FoldTuples
    {
      // add arg to repeated constructor argument list
      template <std::size_t... Is, class... Tuples>
      static Tuple make(Indices<Is...>, Tuples&&... tuples)
      {
        return Tuple{make_element(index_<Is>, std::forward<Tuples>(tuples)...)...};
      }

      template <std::size_t I, class... Tuples>
      static std::tuple_element_t<I, Tuple> make_element(index_t<I>, Tuples&&... tuples)
      {
        using ::AMDiS::get;
        return std::tuple_element_t<I, Tuple>{get<I>(std::forward<Tuples>(tuples))...};
      }
    };

  } // end namespace Impl


  /// Construct a tuple with each element constructed by the same argument arg.
  /// Returns tuple { tuple_element_t<0>(arg), tuple_element_t<1>(arg), ... }
  template <class Tuple, class Arg>
  Tuple constructTuple(Arg&& arg)
  {
    return Impl::ConstructTuple<Tuple, std::tuple_size<Tuple>::value>::make(
        std::forward<Arg>(arg));
  }

  /// Construct a tuple of tuples, by folding a list of tuples of the same size.
  /// Return tuple { {arg0[0], arg1[0], arg2[0], ...}, {arg0[1], arg1[1], arg2[1], ...}, ...}
  template <class Tuple, class... Args>
  Tuple foldTuples(Args&&... args)
  {
    return Impl::FoldTuples<Tuple>::make(MakeSeq_t<std::tuple_size<Tuple>::value>(),
        std::forward<Args>(args)...);
  }


  namespace Impl
  {
    template <template<class> class Base, class Tuple, class Indices> struct MakeTuple;

    template <template<class> class Base, class Tuple, std::size_t... I>
    struct MakeTuple<Base, Tuple, Indices<I...>>
    {
      using type = std::tuple<Base<std::tuple_element_t<I, Tuple>>...>;
    };

  } // end namespace Impl


  /// Constructs tuple type, by wrapping Base around the tuple elements.
  /// Returns tuple type tuple<Base<tuple_element_t<0>>, Base<tuple_element_t<1>>, ...>
  template <template<class> class Base, class Tuple>
  using MakeTuple_t =
    typename Impl::MakeTuple<Base, Tuple, MakeSeq_t<std::tuple_size<Tuple>::value>>::type;


  namespace Impl
  {
    template <template<class,class> class Base, class Tuple1, class Tuple2, class Indices> struct MakeTuple2;

    template <template<class,class> class Base, class Tuple1, class Tuple2, std::size_t... I>
    struct MakeTuple2<Base, Tuple1, Tuple2, Indices<I...>>
    {
      using type = std::tuple<Base<std::tuple_element_t<I, Tuple1>, std::tuple_element_t<I, Tuple2>>...>;
    };

  } // end namespace Impl


  /// Constructs tuple type, by wrapping Base around the tuple elements.
  /// Returns tuple type tuple<Base<tuple_element_t<0,T1>,tuple_element_t<0,T2>>, Base<tuple_element_t<1,T1>,tuple_element_t<1,T2>>, ...>
  template <template<class,class> class Base, class Tuple1, class Tuple2>
  using MakeTuple2_t =
    typename Impl::MakeTuple2<Base, Tuple1, Tuple2, MakeSeq_t<std::tuple_size<Tuple1>::value>>::type;


  namespace Impl
  {
    template <template<class...> class Base, class Tuple, class Indices> struct ExpandTuple;

    template <template<class...> class Base, class Tuple, std::size_t... I>
    struct ExpandTuple<Base, Tuple, Indices<I...>>
    {
      using type = Base<std::tuple_element_t<I, Tuple>...>;
    };

  } // end namespace Impl


  /// Expands tuple element types into template list of Base class.
  /// Returns Base<tuple_element<0>, tuple_element<1>, ...>
  template <template<class...> class Base, class Tuple>
  using ExpandTuple_t =
    typename Impl::ExpandTuple<Base, Tuple, MakeSeq_t<std::tuple_size<Tuple>::value>>::type;


  namespace Impl
  {
    template <class T, class Indices> struct RepeatedTuple;

    template <class T, std::size_t... I>
    struct RepeatedTuple<T, Indices<I...>>
    {
      template <std::size_t, class U> using Id = U;
      using type = std::tuple<Id<I, T>...>;
    };

  } // end namespace Impl

  /// Construct tuple of N times type T.
  /// Returns tuple<T, T, T, ...>
  template <std::size_t N, class T>
  using Repeat_t =
    typename Impl::RepeatedTuple<T, MakeSeq_t<N>>::type;




  namespace Impl
  {
    template <template<class> class Base, class Tuple, class Indices> struct MapTuple;

    template <template<class> class Base, class Tuple, std::size_t... I>
    struct MapTuple<Base, Tuple, Indices<I...>>
    {
      using type = std::tuple<typename Base<std::tuple_element_t<I, Tuple>>::type...>;
    };

  } // end namespace Impl

  /// Constructs tuple type, by wrapping Base around the tuple elements.
  /// Returns tuple type tuple<Base<tuple_element_t<0>>::type, Base<tuple_element_t<1>>::type, ...>
  template <template<class> class Base, class Tuple>
  using MapTuple_t =
    typename Impl::MapTuple<Base, Tuple, MakeSeq_t<std::tuple_size<Tuple>::value>>::type;


  template <class Functor, class Tuples, std::size_t I, std::size_t... J>
  auto mapTuple(Functor f, Tuples&& tuples, index_t<I>, Indices<J...>)
  {
    return f(std::get<I>(std::get<J>(std::forward<Tuples>(tuples)))...);
  }

  template <class Functor, class Tuples, std::size_t... I>
  auto mapTuple(Functor f, Tuples&& tuples, Indices<I...>)
  {
    return std::make_tuple(mapTuple(f, std::forward<Tuples>(tuples), index_<I>,
                                    MakeSeq_t<std::tuple_size<std::decay_t<Tuples>>::value>{})...);
  }

  // construction method to construct a new tuple by applying a functor to the elements
  // of a given tuple
  template <class Functor, class Tuple0, class... Tuples>
  auto mapTuple(Functor f, Tuple0&& tuple0, Tuples&&... tuples)
  {
    return mapTuple(f, std::make_tuple(std::forward<Tuple0>(tuple0), std::forward<Tuples>(tuples)...),
                    MakeSeq_t<std::tuple_size<std::decay_t<Tuple0>>::value>{});
  }


} // end namespace AMDiS
